#
#   Copyright 2020 Joshua Maglione 
#
#   Distributed under MIT License
#

from .Globals import __TIME as _time

# Given a list of posets L, a poset P, and an integer n, decide if L has a poset
# isomorphic to P.
def _check(L, P, n):
    if len(L) == 0 or len(L) <= n:
        return False, None
    if L[n].is_isomorphic(P):
        return True, n
    return _check(L, P, n+1) 


class IADatabase():

    def __init__(self):
        self.poset_list = []
        self.gen_func_list = []

    def __repr__(self):
        return "A database indexed by %s posets" % (len(self.poset_list))

    def has_poset(self, P):
        L = self.poset_list
        return _check(L, P, 0)

    def get_gen_func(self, P, style):
        isit, k = self.has_poset(P)
        if not isit:
            return None
        return self.gen_func_list[k][style]

    def save_gen_func(self, P, style, F):
        assert P.rank() > 2
        assert style in ['Igusa', 'skele']
        isit, k = self.has_poset(P)
        if not isit:
            # New poset, so we save it.
            gen_dict = {
                'Igusa': None,
                'skele': None
            }
            gen_dict[style] = F
            self.poset_list += [P]
            self.gen_func_list += [gen_dict]
        else:
            if self.gen_func_list[k][style] is None:
                GFL = self.gen_func_list
                GFL[k][style] = F
                self.gen_func_list = GFL


def _initialize_main_DB():
    from sage.all import DiGraph, Poset, var
    import hypigu.src.init_data as init_data
    q = var('q')
    t = var('t')
    Y = var('Y')
    T = var('T')
    DB = IADatabase()

    # A3 arrangement 
    A3 = Poset(DiGraph(init_data.A3_rels))
    A3_Igusa = q**-6*(1-q**-1)*(q**4*(6-5*q+q**2)-4*q**4*(2-q)*t-q**2*(3-7*q+2*q**2)*t**2+q**2*(2-7*q+3*q**2)*t**3-4*q*(1-2*q)*t**4-(1-5*q+6*q**2)*t**5)/((1-q**-1*t)**2*(1-q**-2*t**3)*(1-q**-3*t**6))
    A3_skele = ((1 + 6*Y + 11*Y**2 + 6*Y**3) + (11 + 37*Y + 37*Y**2 + 11*Y**3)*T + (6 + 11*Y + 6*Y**2 + Y**3)*T**2)/((1 - T)**3)
    DB.save_gen_func(A3, 'Igusa', A3_Igusa)
    DB.save_gen_func(A3, 'skele', A3_skele)

    # A4 arrangement 
    A4 = Poset(DiGraph(init_data.A4_rels))
    A4_skele = ((1 + 10*Y + 35*Y**2 + 50*Y**3 + 24*Y**4) + (47 + 260*Y + 505*Y**2 + 400*Y**3 + 108*Y**4)*T + (108 + 400*Y + 505*Y**2 + 260*Y**3 + 47*Y**4)*T**2 + (24 + 50*Y + 35*Y**2 + 10*Y**3 + Y**4)*T**3)/((1 - T)**4)
    A4_Igusa = init_data.A4_Igusa_n(q, t) / init_data.A4_Igusa_d(q, t)
    DB.save_gen_func(A4, 'skele', A4_skele)
    DB.save_gen_func(A4, 'Igusa', A4_Igusa)
    
    # A5 arrangement
    A5 = Poset(DiGraph(init_data.A5_rels))
    A5_skele = (Y**5*T**4 + 197*Y**5*T**3 + 15*Y**4*T**4 + 1268*Y**5*T**2 + 1546*Y**4*T**3 + 85*Y**3*T**4 + 1114*Y**5*T + 7172*Y**4*T**2 + 4670*Y**3*T**3 + 225*Y**2*T**4 + 120*Y**5 + 4493*Y**4*T + 15320*Y**3*T**2 + 6700*Y**2*T**3 + 274*Y*T**4 + 274*Y**4 + 6700*Y**3*T + 15320*Y**2*T**2 + 4493*Y*T**3 + 120*T**4 + 225*Y**3 + 4670*Y**2*T + 7172*Y*T**2 + 1114*T**3 + 85*Y**2 + 1546*Y*T + 1268*T**2 + 15*Y + 197*T + 1)/((1 - T)**5)
    A5_Igusa = init_data.A5_Igusa_n(q, t) / init_data.A5_Igusa_d(q, t)
    DB.save_gen_func(A5, 'skele', A5_skele)
    DB.save_gen_func(A5, 'Igusa', A5_Igusa)

    # B3 arrangement
    B3 = Poset(DiGraph(init_data.B3_rels))
    B3_skele = (Y**3*T**2 + 20*Y**3*T + 9*Y**2*T**2 + 15*Y**3 + 76*Y**2*T + 23*Y*T**2 + 23*Y**2 + 76*Y*T + 15*T**2 + 9*Y + 20*T + 1)/((1 - T)**3)
    B3_Igusa = init_data.B3_Igusa_n(q, t) / init_data.B3_Igusa_d(q, t)
    DB.save_gen_func(B3, 'skele', B3_skele)
    DB.save_gen_func(B3, 'Igusa', B3_Igusa)

    return DB

global internal_database
internal_database = _initialize_main_DB()



#
#   Copyright 2020 Joshua Maglione
#
#   Distributed under MIT License
#

from datetime import datetime as _dt
from os import cpu_count as _cpu_count

__PRINT = False
__SANITY = False
__NCPUS = max(1, _cpu_count() - 1)


def __TIME():
    return "[{0}] ".format(_dt.now().strftime("%b %d %H:%M:%S"))



#
#   Copyright 2021 Joshua Maglione
#
#   Distributed under MIT License
#
from sage.all import var

from .Database import internal_database as _data
from .Globals import __PRINT as _print
from .Globals import __TIME as _time
from functools import reduce as _reduce


# A function to return a Poincaré function.
def _Poincare_polynomial(L, sub=None, abs_val=False):
    if sub is None:
        sub = var('Y')

    def poincare(x):
        pi = L.restriction(x).Poincare_polynomial(abs_val=abs_val)
        try:
            return pi.subs({pi.variables()[0]: sub})
        except AttributeError:  # In case pi is a constant.
            return pi
    return poincare


# The complete solutions for small central arrangements of rank <= 2.
def _small_central(A, style, numerator=False):
    p = var('q')
    t = var('t')
    Y = var('Y')
    T = var('T')
    if A.rank() == 1:
        if style == 'Igusa':
            return (1 - p**-1)/(1 - p**-1*t)
        if style == 'skele':
            if numerator:
                return 1 + Y
            else:
                return (1 + Y)/(1 - T)
    # Now we assume the rank == 2.
    m = len(A)
    if style == 'Igusa':
        return (1 - p**-1)*(1 - (m - 1)*p**-1 + (m - 1)*p**-1*t - p**-2*t) / ((1 - p**-1*t)*(1 - p**-2*t**m))
    if style == 'skele':
        if numerator:
            return 1 + m*Y + (m-1)*Y**2 + (m-1 + m*Y + Y**2)*T
        else:
            return (1 + m*Y + (m-1)*Y**2 + (m-1 + m*Y + Y**2)*T)/((1 - T)**2)


# The direct version of the universal generating function computation.
def _universal(L, analytic=False, atom=False):
    from .LatticeFlats import _subposet

    # Set up the potential substitutions for T -- as defined in Maglione--Voll.
    if analytic:
        q = var('q')
        Y = -q**(-1)
        P = L.poset
        t_name = lambda x: var("t" + str(x))
        if atom:
            atoms = P.upper_covers(P.bottom())

            def T_data(x):
                under_poset = _subposet(P, x, P.lower_covers)
                elts = (y for y in atoms if y in under_poset)
                ts = map(t_name, elts)
                return _reduce(lambda x, y: x*y, ts, q**(-P.rank(x)))
        else:
            def T_data(x):
                under_poset = _subposet(P, x, P.lower_covers)
                elts = list(under_poset._elements)
                elts.remove(P.bottom())
                ts = map(t_name, elts)
                return _reduce(lambda x, y: x*y, ts, q**(-P.rank(x)))
    else:
        T_data = lambda x: var("T" + str(x))
        Y = var('Y')

    T = {x: T_data(x) for x in L.poset._elements}

    # Base cases for recursion.
    if L.poset.has_top() and L.poset.rank() == 2:
        elts = L.proper_part_poset()._elements
        merge = lambda x, y: x + (1 + Y)**2*T[y]/(1 - T[y])
        one = L.poset.top()
        return _reduce(merge, elts, (1 + Y)*(1 + (len(elts) - 1)*Y))/(1-T[one])
    if L.poset.rank == 1:
        elts = list(L.poset._elements).remove(L.poset.bottom())
        merge = lambda x, y: x + (1 + Y)*T[y]/(1 - T[y])
        return _reduce(merge, elts, 1 + len(elts)*Y)

    P = L.proper_part_poset()
    poincare = _Poincare_polynomial(L, sub=Y)
    recurse = lambda M: _universal(M, analytic=analytic, atom=atom)
    num_dat = lambda x: poincare(x)*T[x]*recurse(L.subarrangement(x))
    factors = map(num_dat, P._elements)
    HP = _reduce(lambda x, y: x + y, factors, poincare(L.poset.bottom()))
    if L.poset.has_top():
        HP = HP/(1 - T[L.poset.top()])
    return HP


def _Igusa_zeta_function(L, DB=True, verbose=_print):
    from .Constructors import CoxeterArrangement
    from .Braid import BraidArrangementIgusa
    from .LatticeFlats import LatticeOfFlats, _Coxeter_poset_data

    P = L.poset
    q = var('q')
    t = var('t')

    # Base cases for recursion.
    if P.has_top() and P.rank() == 2:
        m = len(P) - 2
        return (1 - q**-1)*(1 - (m-1)*q**-1 + m*(1 - q**-1)*q**-1*t/(1 - q**-1*t))/(1 - q**-2*t**m)
    if P.rank() == 1:
        m = len(P) - 1
        return 1 - m*q**-1 + m*(1 - q**-1)*q**-1*t/(1 - q**-1*t)

    if DB:
        zeta = _data.get_gen_func(P, 'Igusa')
        if zeta is not None:
            return zeta
    # We check to see if we have a type A braid arrangement.
    # We can compute these *extremely* quickly.
    if _Coxeter_poset_data()['A']['hyperplanes'](P.rank()) == len(L.atoms()):
        if _Coxeter_poset_data()['A']['poset'](P.rank()) == len(P):
            B = CoxeterArrangement("A" + str(P.rank()))
            if P.is_isomorphic(LatticeOfFlats(B).poset):
                return BraidArrangementIgusa(P.rank())

    poincare = _Poincare_polynomial(L, sub=-q**(-1))
    t_factor = lambda X: t**len(L.flat_labels[X])
    x_factor = lambda x: poincare(x)*t_factor(x)*q**(-P.rank(x))
    eq_elt_data = L._combinatorial_eq_elts()
    factors = map(lambda x: x[1]*x_factor(x[0]), eq_elt_data)
    integrals = map(lambda x: _Igusa_zeta_function(x[2], DB=DB), eq_elt_data)
    pi = poincare(P.bottom())
    zeta = _reduce(lambda x, y: x + y[0]*y[1], zip(factors, integrals), 0) + pi
    if P.has_top():
        zeta = zeta/(1 - q**(-P.rank())*t**len(L.atoms()))
    if DB and P.rank() > 2:
        _data.save_gen_func(P, 'Igusa', zeta)
    return zeta


def _top_zeta_function_uni(L, DB=True, verbose=_print):
    P = L.poset
    s = var('s')
    C = 1*L.poset.has_top()

    # Base cases for recursion.
    if P.has_top() and P.rank() == 2:
        m = len(P) - 2
        return (2 + (2 - m)*s)/((2 + m*s)*(1 + s))
    if P.rank() == 1:
        m = len(P) - 1
        return (1 + (1 - m)*s)/(1 + s)

    poincare = _Poincare_polynomial(L)
    Y = poincare(P.bottom()).variables()[0]
    pi_circ = lambda x: (poincare(x)/(1 + Y)**C).factor().simplify().subs({Y: -1})
    eq_elt_data = L._combinatorial_eq_elts()
    factors = map(lambda x: x[1]*pi_circ(x[0]), eq_elt_data)
    integrals = map(lambda x: _top_zeta_function_uni(x[2], DB=DB), eq_elt_data)
    pi = pi_circ(P.bottom())
    zeta = _reduce(lambda x, y: x + y[0]*y[1], zip(factors, integrals), 0) + pi
    if C == 1:
        zeta = zeta/(P.rank() + len(L.atoms())*s)

    return zeta


def _top_zeta_function_mul(L, DB=True, verbose=_print, atom=False):
    from .LatticeFlats import _subposet

    P = L.poset
    C = 1 * L.poset.has_top()

    s_name = lambda x: var("s" + str(x))
    if atom:
        if atom:
            atoms = P.upper_covers(P.bottom())

            def s_data(x):
                under_poset = _subposet(P, x, P.lower_covers)
                elts = (y for y in atoms if y in under_poset)
                ts = map(s_name, elts)
                return _reduce(lambda x, y: x + y, ts, 0)
    else:
        def s_data(x):
            under_poset = _subposet(P, x, P.lower_covers)
            elts = list(under_poset._elements)
            elts.remove(P.bottom())
            ts = map(s_name, elts)
            return _reduce(lambda x, y: x + y, ts, 0)

    S = {x: s_data(x) for x in P._elements}

    # Base cases for recursion.
    add_em = lambda x, y: x + y
    if P.has_top() and P.rank() == 2:
        atms = P.upper_covers(P.bottom())
        m = len(atms)
        elt_dat = lambda x: 1/(1 + S[x])
        return _reduce(add_em, map(elt_dat, atms), 2 - m)/(2 + S[P.top()])
    if P.rank() == 1:
        atms = P.upper_covers(P.bottom())
        m = len(atms)
        elt_dat = lambda x: 1/(1 + S[x])
        return _reduce(add_em, map(elt_dat, atms), 1 - m)

    poincare = _Poincare_polynomial(L)
    Y = poincare(P.bottom()).variables()[0]
    pi_circ = lambda x: (poincare(x)/(1 + Y)**C).factor().simplify().subs({Y: -1})
    x_factor = pi_circ
    prop_elts = L.proper_part_poset()._elements
    factors = map(lambda x: x_factor(x), prop_elts)
    integrals = map(lambda x: _top_zeta_function_mul(L.subarrangement(x), DB=DB, atom=atom), prop_elts)
    pi = pi_circ(P.bottom())
    zeta = _reduce(lambda x, y: x + y[0]*y[1], zip(factors, integrals), 0) + pi
    if P.has_top():
        zeta = zeta/(P.rank() + S[P.top()])

    return zeta


def _comb_skele(L, abs_val=False, DB=True, verbose=_print):
    P = L.poset
    Y = var('Y')
    T = var('T')

    if P.rank() == 1:
        a = len(L.atoms())
        return (1 + a*Y)/(1 - T)
    if P.rank() == 2 and P.has_top():
        m = len(P) - 2
        return (1 + m*Y + (m - 1)*Y**2 + (m - 1 + m*Y + Y**2)*T)/(1 - T)**2
    if DB:
        if verbose:
            print(_time() + "Checking database.")
        zeta = _data.get_gen_func(P, 'skele')
        if zeta is not None:
            if verbose:
                print("\tFound it.")
            return zeta
        else:
            if verbose:
                print("\tDid not find it.")

    poincare = _Poincare_polynomial(L, abs_val=abs_val)
    if verbose:
        print(_time() + "Gleaning structure from poset.")
    eq_elt_data = L._combinatorial_eq_elts()
    if verbose:
        print("\tDone.")
        print(_time() + "Lattice points: {0},  Relevant points: {1}".format(len(P), len(eq_elt_data)))
    factors = map(lambda x: x[1]*T*poincare(x[0]), eq_elt_data)
    if verbose:
        print(_time() + "Recursing...")
    integrals = map(lambda x: _comb_skele(x[2], DB=DB, abs_val=abs_val, verbose=verbose), eq_elt_data)
    if verbose:
        print(_time() + "Putting everything together...")
    pi = poincare(P.bottom())
    zeta = _reduce(lambda x, y: x + y[0]*y[1], zip(factors, integrals), 0) + pi
    if P.has_top():
        zeta = zeta/(1 - T)
    if DB and P.rank() > 2:
        _data.save_gen_func(P, 'skele', zeta)

    return zeta


# Given a polynomial, return a hyperplane arrangement equivalent to the linear
# factors of f.
def _parse_poly(f):
    from sage.all import SR, QQ, HyperplaneArrangements, Matrix
    if isinstance(f, str):
        f = SR(f)
    if f.base_ring() == SR:
        L = f.factor_list()
        K = QQ
    else:
        L = list(f.factor())
        K = f.base_ring()

    L = filter(lambda T: not T[0] in K, L)  # Remove constant factors
    F, M = list(zip(*L))

    # Verify that each polynomial factor is linear
    is_lin = lambda g: all(g.degree(x) <= 1 for x in g.variables())
    if not all(map(is_lin, F)):
        raise ValueError("Expected product of linear factors.")

    varbs = f.variables()
    varbs_str = tuple(str(x) for x in varbs)
    HH = HyperplaneArrangements(K, varbs_str)

    def poly_vec(g):
        c = K(g.subs({x: 0 for x in g.variables()}))
        return tuple([c] + [K(g.coefficient(x)) for x in varbs])

    F_vec = tuple(map(poly_vec, F))
    A = HH(Matrix(K, F_vec))

    # This scrambles the hyperplanes, so we need to scramble M in the same way.
    A_vec = tuple(map(lambda H: tuple(H.coefficients()), A.hyperplanes()))
    perm = (F_vec.index(v) for v in A_vec)
    M_new = tuple([M[i] for i in perm])

    return A, M_new


def CoarseFlagHPSeries(A=None, lattice_of_flats=None, poset=None, matroid=None, numerator=False, abs_val=False, verbose=_print):
    from .LatticeFlats import LatticeOfFlats

    if lattice_of_flats is None:
        if verbose:
            print("{0}Building lattice of flats".format(_time()))
        if matroid is None and poset is None:
            L = LatticeOfFlats(A=A)
        else:
            if poset is None:
                L = LatticeOfFlats(matroid=matroid)
            else:
                L = LatticeOfFlats(poset=poset)
    else:
        L = lattice_of_flats

    if verbose:
        print("{0}Computing coarse flag Hilbert--Poincare series".format(_time()))
    cfHP = _comb_skele(L, abs_val=abs_val, verbose=verbose)

    if numerator:
        if verbose:
            print(_time() + "Constructing numerator.")
        D = cfHP.numerator_denominator()[1]
        T = D.variables()[0]
        if D == (T - 1)**(L.poset.rank()):
            e = -1
        if D == (1 - T)**(L.poset.rank()):
            e = 1
        return e*(cfHP*D).factor()
    else:
        return cfHP


def IgusaZetaFunction(X=None, lattice_of_flats=None, int_poset=None, matroid=None, verbose=_print):
    from .LatticeFlats import LatticeOfFlats

    HPA = True
    if matroid is None:
        try:
            # Check if a hyperplane arrangement.
            _ = X.hyperplanes()
            A = X
        except AttributeError:
            # Not an HPA; deal with polynomial input.
            A, M = _parse_poly(X)
            if verbose:
                print("{0}Constructed a hyperplane arrangement".format(_time()))
            HPA = False

    if lattice_of_flats is None:
        if verbose:
            print("{0}Building lattice of flats".format(_time()))
        if matroid is None:
            L = LatticeOfFlats(A, poset=int_poset)
        else:
            L = LatticeOfFlats(matroid=matroid)
    else:
        L = lattice_of_flats

    if not HPA:
        if list(M) == [1]*len(M):
            if verbose:
                print("{0}Computing Igusa's zeta function".format(_time()))
            return _Igusa_zeta_function(L)
        else:
            if verbose:
                print("{0}Computing the atom zeta function".format(_time()))
            Z = _universal(L, analytic=True, atom=True)
            t = var('t')
            SUB = {var('t' + str(k+1)): t**(M[k]) for k in range(len(M))}
            return Z.subs(SUB)

    if verbose:
        print("{0}Computing Igusa's zeta function".format(_time()))
    return _Igusa_zeta_function(L)


def TopologicalZetaFunction(X=None, lattice_of_flats=None, int_poset=None, verbose=_print, multivariate=False, atom=False, matroid=None):
    from .LatticeFlats import LatticeOfFlats

    HPA = True
    if matroid is None:
        try:
            # Check if a hyperplane arrangement.
            _ = X.hyperplanes()
            A = X
        except AttributeError:
            # Not an HPA; deal with polynomial input.
            A, M = _parse_poly(X)
            if verbose:
                print("{0}Constructed a hyperplane arrangement".format(_time()))
            HPA = False

    if lattice_of_flats is None:
        if matroid is None:
            if verbose:
                print("{0}Building lattice of flats".format(_time()))
            L = LatticeOfFlats(A, poset=int_poset)
        else:
            L = LatticeOfFlats(matroid=matroid)
    else:
        L = lattice_of_flats

    if verbose:
        print("{0}Computing the topological zeta function".format(_time()))

    if not HPA:
        if all(m == 1 for m in M):
            return _top_zeta_function_uni(L)
        else:
            Z = _top_zeta_function_mul(L, atom=True)
            s = var('s')
            SUB = {var('s' + str(k+1)): M[k]*s for k in range(len(M))}
            return Z.subs(SUB)

    if not multivariate:
        return _top_zeta_function_uni(L)

    return _top_zeta_function_mul(L, atom=atom)


def AnalyticZetaFunction(A=None, lattice_of_flats=None, int_poset=None, matroid=None, verbose=_print):
    from .LatticeFlats import LatticeOfFlats

    if lattice_of_flats is None:
        if matroid is None:
            if verbose:
                print("{0}Building lattice of flats".format(_time()))
            L = LatticeOfFlats(A, poset=int_poset)
        else:
            L = LatticeOfFlats(matroid=matroid)
    else:
        L = lattice_of_flats

    if verbose:
        print("{0}Computing the analytic zeta function".format(_time()))
    return _universal(L, analytic=True)


def AtomZetaFunction(A=None, lattice_of_flats=None, int_poset=None, matroid=None, verbose=_print):
    from .LatticeFlats import LatticeOfFlats

    if lattice_of_flats is None:
        if matroid is None:
            if verbose:
                print("{0}Building lattice of flats".format(_time()))
            L = LatticeOfFlats(A, poset=int_poset)
        else:
            L = LatticeOfFlats(matroid=matroid)
    else:
        L = lattice_of_flats

    if verbose:
        print("{0}Computing the atom zeta function".format(_time()))
    return _universal(L, analytic=True, atom=True)


def FlagHilbertPoincareSeries(A=None, lattice_of_flats=None, int_poset=None, matroid=None, verbose=_print):
    from .LatticeFlats import LatticeOfFlats

    if lattice_of_flats is None:
        if matroid is None:
            if verbose:
                print("{0}Building lattice of flats".format(_time()))
            L = LatticeOfFlats(A, poset=int_poset)
        else:
            L = LatticeOfFlats(matroid=matroid)
    else:
        L = lattice_of_flats

    if verbose:
        print("{0}Computing the flag Hilbert--Poincaré series".format(_time()))
    return _universal(L)



#
#   Copyright 2020 Joshua Maglione
#
#   Distributed under MIT License
#

from functools import reduce
from sage.misc.cachefunc import cached_method
from .Globals import __TIME as _time
from .Globals import __NCPUS as _N
import sage.parallel.decorate as _para


def _contract(M, rows):
    from sage.all import Matrix
    K = M.base_ring()
    Q = [M[k] for k in rows] + [M[k] for k in range(M.nrows()) if k not in rows]
    Q = Matrix(K, Q).transpose()
    top = Q[0]
    E = Q[1:, :].echelon_form()
    rel_E = E[:, :len(rows)]
    out_E = E[:, len(rows):]
    rm = rel_E.pivot_rows()
    # In case we have a non-central arrangement
    for r in rm:
        v = E[r]
        i = list(v).index(1)
        top = top - top[i]*v
    A = [tuple(list(top)[len(rows):])] + [out_E[k] for k in range(E.nrows())
                                          if k not in rm]
    M_out = Matrix(K, A).transpose()
    return M_out


# BUILD A WAY TO GET THE LABELS FROM THE CONTRACT MATRIX
def _get_labels(M, x, rows, L):
    from sage.all import VectorSpace, Set, Matrix

    # If non-central, then this is true iff not intersecting.
    not_e1 = lambda v: list(v)[1:] != [0]*(len(v)-1)

    # Determine new hyperplanes and group like rows together.
    V = VectorSpace(M.base_ring(), M.ncols())
    lines = []
    labels = []
    for r in range(M.nrows()):
        v = M[r]
        is_new = not_e1(v)
        i = 0
        while i < len(lines) and is_new:
            if v in V.subspace([lines[i]]):
                is_new = False
                labels[i] = labels[i].union(Set([r]))
            else:
                i += 1
        if is_new:
            lines.append(v)
            labels.append(Set([r]))

    # Adjust the labels because we are missing rows.
    fix_sets = lambda F: lambda S: Set(list(map(lambda s: F(s), S)))
    for k in rows:
        adjust = lambda i: i + (k <= i)*1
        labels = list(map(fix_sets(adjust), labels))
    labels = [Set(rows)] + labels

    # Adjust the row labels to hyperplane labels
    HL = L.hyperplane_labels
    A = L.hyperplane_arrangement
    HL_lab = lambda i: list(filter(lambda j: HL[j] == A[i], HL.keys()))[0]
    labels = list(map(fix_sets(HL_lab), labels))

    # Get the new hyperplanes
    FL = L.flat_labels
    new_hyp = [labels[k].union(labels[0]) for k in range(1, len(labels))]
    P = L.poset
    flat = lambda S: list(filter(lambda j: FL[j] == S, P.upper_covers(x)))[0]
    new_hyp = list(map(flat, new_hyp))

    def last_adj(S):
        l_hyp = labels[1:]
        T = S.difference(labels[0])
        new_T = Set([])
        for k in range(len(l_hyp)):
            if l_hyp[k].issubset(T):
                new_T = new_T.union(Set([new_hyp[k]]))
        return new_T

    return Matrix(lines), last_adj, {new_hyp[i]: i for i in range(len(new_hyp))}


def _parse_poset(P):
    global POS, atoms, labs, int_at
    from sage.all import Set
    import sage.parallel.decorate as para

    POS = P
    N = _N
    atoms = POS.upper_covers(POS.bottom())

    @para.parallel(N)
    def atom_set(k, shift):
        S = list(POS._elements[1+shift::k])
        m = lambda x: [x, Set(filter(lambda a: POS.le(a, x), atoms))]
        return list(map(m, S))

    labs = list(atom_set([(N, k) for k in range(N)]))
    labs = [[P.bottom(), Set([])]] + reduce(lambda x, y: x + y[1], labs, [])
    label_dict = {T[0]: T[1] for T in labs}

    return label_dict


# Can return the subarrangement A_x or restriction A^x simply based on the
# function F given. For subarrangement use 'P.lower_covers' and for
# restriction use 'P.upper_covers'.
def _subposet(P, x, F):
    from sage.all import Set
    elts = Set()
    new_level = Set([x])
    while len(elts.union(new_level)) > len(elts):
        elts = elts.union(new_level)
        new_level = Set(reduce(
            lambda x, y: x+y,
            map(F, new_level),
            []
        ))
    return P.subposet(elts)


# Parallel function to build the intersection lattice.
# Moved to global to prevent accidentally carrying unnecessary data.
@_para.parallel(_N)
def build_next(A, S, HYP, LIN):
    from sage.all import exists, Set
    new_level = []
    new_hypcont = []
    if len(S) > 0:
        m = S[0]
    for i in S:
        T = LIN[i - m]
        for j in range(len(A)):
            # Skip the hyperplane already known to contain the intersection.
            if j not in HYP[i - m]:
                H = A[j]
                I = H._affine_subspace().intersection(T)
                # Check if the intersection is trivial.
                if I is not None:
                    if I == T:
                        # This case means that H cap T = T, so we should
                        # record that H contains T.
                        HYP[i - m] = HYP[i - m].union(Set([j]))
                    else:
                        # Check if we have this intersection already.
                        is_in, ind = exists(
                            range(len(new_level)),
                            lambda k: I == new_level[k]
                        )
                        if is_in:
                            # We have the intersection, so we update
                            # containment info accordingly.
                            new_hypcont[ind] = new_hypcont[ind].union(
                                Set([j]).union(HYP[i - m])
                            )
                        else:
                            # We do not have it, so we update everything.
                            new_level.append(I)
                            new_hypcont.append(HYP[i - m].union(Set([j])))
    return list(zip(new_level, new_hypcont))


# Default SageMath algorithm works well. However 'A.matroid()' seems to remove
# ordering, which we depend on, so care is needed.
def _lof_from_matroid(A=None, matroid=None):
    from sage.all import Set, Matrix, Matroid, Poset
    if A is not None:
        rows = list(map(lambda H: H.coefficients()[1:], A.hyperplanes()))
        mat = Matrix(A.base_ring(), rows).transpose()
        M = Matroid(mat)
        n = len(A)
        lbl_map = lambda S: S
    else:
        M = matroid.simplify()
        n = len(M.groundset())
        grd_list = list(M.groundset())
        l_map = {x: grd_list.index(x) for x in grd_list}
        lbl_map = lambda S: frozenset([l_map[x] for x in S])

    L = M.lattice_of_flats()
    rank_r = lambda L, r: list(
        map(lbl_map, filter(lambda x: L.rank(x) == r, L._elements))
    )
    rank_1 = [frozenset([k]) for k in range(n)]
    ranks = reduce(
        lambda x, y: x + y,
        [rank_r(L, r) for r in range(2, L.rank() + 1)],
        [L.bottom()] + rank_1
    )
    P = Poset(
        L, element_labels={x: ranks.index(lbl_map(x)) for x in L._elements}
    )
    adj_set = lambda S: Set([x+1 for x in S])
    label_dict = {i: adj_set(ranks[i]) for i in range(len(L))}
    if A is not None:
        hyp_dict = {i: A[list(ranks[i])[0]] for i in range(1, n + 1)}
    else:
        hyp_dict = None
    return [P, label_dict, hyp_dict]


def _lof_from_affine_matroid(A):
    from sage.all import Poset, Set
    A_coned = A.cone()
    hyps = list(map(lambda H: H.coefficients(), A_coned.hyperplanes()))
    extra = [0, 1] + [0]*(A.dimension())
    i = hyps.index(extra)
    P, L, H = _lof_from_matroid(A_coned)
    assert (H[i + 1]).coefficients() == extra
    new_elts = list(filter(lambda x: not P.le(i + 1, x), P))
    new_elts = reduce(
        lambda x, y: x + y,
        [list(filter(lambda x: P.rank(x) == r, new_elts)) for r in range(2, P.rank() + 1)],
        [0] + [j for j in range(1, len(A_coned) + 1) if j != i + 1]
    )
    new_names = {x: new_elts.index(x) for x in new_elts}
    adj_set = lambda S: Set([new_names[x] for x in S if x in new_elts])
    P_new = Poset(P.subposet(new_elts), element_labels=new_names)
    L_new = {new_names[x]: adj_set(L[x]) for x in new_elts}

    def inv_H(h):
        cut = lambda x: x.coefficients()[1:]
        pair = list(filter(lambda x: cut(x[1]) == h, H.items()))
        return pair[0][0]
    H_new = {new_names[inv_H(h.coefficients())]: h for h in A}
    return [P_new, L_new, H_new]


class LatticeOfFlats():

    def __init__(self, A=None, poset=None, flat_labels=None,
    hyperplane_labels=None, lazy=False, matroid=None,
    nature_hyperplane_label=True):
        self.hyperplane_arrangement = A
        self.poset = poset
        self.flat_labels = flat_labels
        self.hyperplane_labels = hyperplane_labels
        if poset is not None:
            assert poset.has_bottom(), "Expected a unique minimal element in poset."
            assert poset.is_graded(), "Expected a graded poset."
            self.poset = poset
        else:
            if not lazy:
                if A is not None:
                    if A.is_central():
                        P, FL, HL = _lof_from_matroid(A)
                    else:
                        P, FL, HL = _lof_from_affine_matroid(A)
                else:
                    P, FL, HL = _lof_from_matroid(A=None, matroid=matroid)
                self.poset = P
                self.flat_labels = FL
                self.hyperplane_labels = HL
        if self.flat_labels is None and not lazy:
            self.flat_labels = _parse_poset(poset)
        if self.hyperplane_arrangement is not None and self.hyperplane_labels is None and nature_hyperplane_label:
            self.hyperplane_labels = {i + 1: A[i] for i in range(len(A))}

    def __repr__(self):
        if self.hyperplane_arrangement:
            return "The lattice of flats of:\n{0}\ngiven by:\n{1}".format(self.hyperplane_arrangement, self.poset)
        else:
            return "The lattice of flats of some matroid given by:\n{0}".format(self.poset)

    def _save(self, file, var_name='L'):
        from sage.all import Matrix
        HH = self.hyperplane_arrangement.parent()
        A = Matrix(map(lambda H: H.coefficients(), self.hyperplane_arrangement.hyperplanes())).rows()
        CR = tuple(map(lambda T: tuple(T), self.poset.cover_relations()))
        FL = self.flat_labels
        FL_tup = tuple([tuple([x, list(FL[x])]) for x in FL.keys()])
        del FL
        dict_builder = "FL = {x[0]: Set(x[1]) for x in FL_tup}\n"
        with open(file, "w") as F:
            F.write("from sage.all import HyperplaneArrangements, QQ, Poset, Set\n")
            F.write("import hypigu as hi\n")
            F.write("H = HyperplaneArrangements(QQ, {0})\n".format(HH.variable_names()))
            del HH
            F.write("A = H({0})\n".format(A).replace("), ", "),\n"))
            del A
            F.write("CR = {0}\n".format(CR).replace("), ", "),\n"))
            del CR
            F.write("P = Poset([range({0}), CR], cover_relations=True)\n".format(len(self.poset._elements)))
            F.write("FL_tup = {0}\n".format(FL_tup).replace("), ", "),\n"))
            F.write(dict_builder)
            F.write("del FL_tup\n")
            F.write("{0} = hi.LatticeOfFlats(A, poset=P, flat_labels=FL)\n".format(var_name))
            F.write("del H, A, CR, P, FL\n")
            F.write("print('Loaded a lattice of flats. Variable name: {0}')".format(var_name))

    def atoms(self):
        return self.poset.upper_covers(self.poset.bottom())

    def labels_of_flats(self):
        elt_tup = lambda x: tuple([x, self.flat_labels[x]])
        return list(map(elt_tup, self.poset._elements))

    def labels_of_hyperplanes(self):
        P = self.poset
        elt_tup = lambda x: tuple([x, self.hyperplane_labels[x]])
        return list(map(elt_tup, P.upper_covers(P.bottom())))

    def proper_part_poset(self):
        P = self.poset
        elts = list(P._elements)
        if P.has_top():
            elts.remove(P.top())
        elts.remove(P.bottom())
        return P.subposet(elts)

    def show(self):
        self.poset.show()

    def subarrangement(self, x):
        P = self.poset
        if not isinstance(x, set):
            assert x in P, "Expected element to be in poset."
            new_P = _subposet(P, x, P.lower_covers)
            new_A = None
            new_FL = None
            new_HL = None
            if self.hyperplane_arrangement and self.hyperplane_labels:
                A = self.hyperplane_arrangement
                HL = self.hyperplane_labels
                atoms = new_P.upper_covers(new_P.bottom())
                keep = list(map(lambda k: HL[k], atoms))
                new_A = A.parent()(keep)
                new_HL = {a: HL[a] for a in atoms}
            if self.flat_labels:
                FL = self.flat_labels
                new_FL = {x: FL[x] for x in new_P._elements}
            return LatticeOfFlats(new_A, poset=new_P, flat_labels=new_FL, hyperplane_labels=new_HL)
        else:
            L = self.flat_labels
            X = list(filter(lambda y: L[y] == x, P._elements))
            try:
                return self.subarrangement(X[0])
            except IndexError:
                raise ValueError("No element labeled by:\n{0}".format(x))

    def restriction(self, x):
        from sage.all import Matrix, HyperplaneArrangements
        P = self.poset
        if not isinstance(x, set):
            assert x in P, "Expected element to be in poset."
            new_P = _subposet(P, x, P.upper_covers)
            new_A = None
            new_HL = None
            if self.hyperplane_arrangement:
                A = self.hyperplane_arrangement
                hyp_coeffs = map(lambda H: H.coefficients(), A.hyperplanes())
                M = Matrix(A.base_ring(), list(hyp_coeffs))
                rows = sorted(list(map(
                    lambda H: list(A).index(self.hyperplane_labels[H]),
                    self.flat_labels[x]
                )))
                new_M = _contract(M, rows)
                new_M, lab_func, hyp_dict = _get_labels(new_M, x, rows, self)
                HH = HyperplaneArrangements(
                    A.base_ring(),
                    A.parent().variable_names()[:new_M.ncols()-1]
                )
                new_A = HH(new_M)
                FL = self.flat_labels
                new_FL = {x: lab_func(FL[x]) for x in new_P._elements}
                new_HL = {a: new_A[hyp_dict[a]] for a in new_P.upper_covers(new_P.bottom())}
            else:
                FL = self.flat_labels
                new_FL = {y: FL[y].difference(FL[x]) for y in new_P._elements}
            return LatticeOfFlats(new_A, poset=new_P, flat_labels=new_FL, hyperplane_labels=new_HL)
        else:
            L = self.flat_labels
            X = list(filter(lambda y: L[y] == x, P._elements))
            try:
                return self.restriction(X[0])
            except IndexError:
                raise ValueError("No element labeled by:\n{0}".format(x))

    def deletion(self, H):
        from sage.all import Set

        P = self.poset
        L = self.flat_labels

        if isinstance(H, set):
            L = self.flat_labels
            X = [y for y in P._elements if L[y] == H]
            try:
                return self.deletion(X[0])
            except IndexError:
                raise ValueError("No element labeled by:\n{0}".format(H))

        assert P.rank_function()(H) == 1, "Expected an atom."

        if P.has_top():
            coatoms = P.lower_covers(P.top())
        else:
            # not really coatoms... but whatever
            coatoms = P.maximal_elements()

        m = len(self.atoms()) - 1

        def check(C):
            S = L[C]
            return len(S) == m and H not in S

        new_top = list(filter(check, coatoms))

        if len(new_top) == 1:
            new_P = _subposet(P, new_top[0], P.lower_covers)
            new_FL = {y: L[y] for y in new_P._elements}
        else:
            def good_flats(F):
                S = L[F]
                if H in S:
                    U = S.difference(Set([H]))
                    return (U != 0) and (U not in L.values())
                else:
                    return True
            flats = list(filter(good_flats, P._elements))
            new_P = P.subposet(flats)
            new_FL = {y: L[y].difference(Set([H])) for y in flats}

        if self.hyperplane_arrangement:
            HPA = self.hyperplane_arrangement
            HL = self.hyperplane_labels
            A = list(HPA)
            A.remove(HPA[H])
            new_HPA = HPA.parent()(A)
            new_HL = {x: HL[x] for x in new_P.upper_covers(new_P.bottom())}
        else:
            new_HPA = None
            new_HL = None

        return LatticeOfFlats(new_HPA, poset=new_P, flat_labels=new_FL, hyperplane_labels=new_HL)

    def _lazy_restriction(self, H):
        HPA = self.hyperplane_arrangement
        assert HPA is not None, "Needs underlying hyperplane arrangement."
        A = HPA.restriction(HPA[H - 1])
        return LatticeOfFlats(A, lazy=True)

    def _lazy_deletion(self, H):
        HPA = self.hyperplane_arrangement
        assert HPA is not None, "Needs underlying hyperplane arrangement."
        return LatticeOfFlats(HPA.parent()(HPA[:H-1] + HPA[H:]), lazy=True)

    @cached_method
    def Poincare_polynomial(self, abs_val=False):
        from sage.all import QQ, PolynomialRing
        if abs_val not in {True, False}:
            raise ValueError("abs must be either True or False.")
        PR = PolynomialRing(QQ, 'Y')
        Y = PR.gens()[0]
        if self.poset is not None:
            P = self.poset
            atoms = self.atoms()
            if P.rank() == 0:
                return PR(1)
            if P.rank() == 1:
                return PR(1 + len(atoms)*Y)
        else:
            # Lazy
            A = self.hyperplane_arrangement
            assert A is not None, "Expected either a poset or hyperplane arrangement."
            if A.rank() == 0:
                return PR(1)
            if A.rank() == 1:
                return PR(1 + len(A)*Y)
        if self.hyperplane_arrangement is not None:
            try:  # Some hyperplane arrangements are bugged in SageMath.
                D = self._lazy_deletion(1)
                R = self._lazy_restriction(1)
                return PR(D.Poincare_polynomial() + Y*R.Poincare_polynomial())
            except Exception:
                pass
        if abs_val:
            comb_elts = self._combinatorial_eq_elts()
            rk = P.rank_function()
            bot = P.bottom()
            pi = PR(1)
            if P.has_top():
                top = P.top()
                pi += abs(P.moebius_function(bot, top))*Y**(rk(top))
            for t in comb_elts:
                pi += abs(P.moebius_function(bot, t[0]))*t[1]*Y**(rk(t[0]))
            return pi
        else:
            chi = P.characteristic_polynomial()
            q = chi.variables()[0]
            d = chi.degree(q)
            return PR((-Y)**d*chi.subs({q: -Y**-1}))

    @cached_method
    def _combinatorial_eq_elts(self):
        global POS, P_elts
        import sage.parallel.decorate as para

        N = _N
        POS = self.poset
        P_elts = self.proper_part_poset()._elements

        @para.parallel(N)
        def match_elts(k, shift):
            all_elts = P_elts[shift::k]
            eq_elts = []
            counts = []
            down = []
            restrict = []
            while len(all_elts) > 0:
                x = all_elts[0]
                dow_x = self.subarrangement(x)
                res_x = self.restriction(x)
                match = False
                i = 0
                while not match and i < len(eq_elts):
                    if dow_x.poset.is_isomorphic(down[i].poset) and res_x.poset.is_isomorphic(restrict[i].poset):
                        match = True
                    else:
                        i += 1
                if match:
                    counts[i] += 1
                else:
                    eq_elts.append(x)
                    counts.append(1)
                    down.append(dow_x)
                    restrict.append(res_x)
                all_elts = all_elts[1:]
            return list(zip(eq_elts, counts, down, restrict))

        # Get the preliminary set of inequivalent elements
        prelim_elts = list(match_elts([(N, k) for k in range(N)]))
        prelim_elts = reduce(lambda x, y: x + y[1], prelim_elts, [])

        # Test further to minimize the size.
        equiv_elts = []
        while prelim_elts:
            x = prelim_elts[0]
            match = False
            i = 0
            while not match and i < len(equiv_elts):
                if x[2].poset.is_isomorphic(equiv_elts[i][2].poset) and x[3].poset.is_isomorphic(equiv_elts[i][3].poset):
                    match = True
                else:
                    i += 1
            if match:
                equiv_elts[i][1] += x[1]
            else:
                equiv_elts.append(list(x))
            prelim_elts = prelim_elts[1:]
        return equiv_elts



#
#   Copyright 2020 Joshua Maglione
#
#   Distributed under MIT License
#

from sage.all import binomial as _binomial
from sage.all import factorial as _factorial
from functools import reduce as _reduce

_TABLE_CUTOFF = 3


def _Igusa_braid_table(p, t, n, style="standard"):
    if n <= 0:
        return 1
    if n == 1:
        if style == "reduced":
            return 2/(1 - t)
        elif style == "skeleton":
            return (1 + p)/(1 - t)
        else:
            return (1 - p**-1)/(1 - p**-1*t)
    if n == 2:
        if style == "reduced":
            return (6 + 6*t)/((1 - t)*(1 - t**3))
        elif style == "skeleton":
            return ((1 + 3*p + 2*p**2) + (2 + 3*p + p**2)*t)/((1 - t)**2)
        else:
            return (1-p**-1)*(1-2*p**-1+2*p**-1*t-p**-2*t)/((1-p**-1*t)*(1-p**-2*t**3))
    if n == 3:
        if style == "reduced":
            return (24 + 48*t + 24*t**2 + 48*t**3 + 24*t**4)/((1 - t)*(1 - t**3)*(1 - t**6))
        elif style == "skeleton":
            return ((1 + 6*p + 11*p**2 + 6*p**3) + (11 + 37*p + 37*p**2 + 11*p**3)*t + (6 + 11*p + 6*p**2 + p**3)*t**2)/((1 - t)**3)
        else:
            return p**-6*(1-p**-1)*(p**4*(6-5*p+p**2)-4*p**4*(2-p)*t-p**2*(3-7*p+2*p**2)*t**2+p**2*(2-7*p+3*p**2)*t**3-4*p*(1-2*p)*t**4-(1-5*p+6*p**2)*t**5)/((1-p**-1*t)**2*(1-p**-2*t**3)*(1-p**-3*t**6))
    raise ValueError("Bad n-value.")


# Counts the number of set partitions of [L[0] + ... + L[-1]]. So instead of
# running through all *labeled* partitions, we just run through the partition
# "shape" and count the number of different labelings.
def _P(L):
    from sage.all import Set
    # Compute binom(n, p_1)*binom(n - p_1, p_2)*binom(n - p_1 - p_2, p_3)*...

    def binom(n, P):
        if len(P) == 1:
            return _binomial(n, P[0])
        else:
            return _binomial(n, P[0]) * binom(n - P[0], P[1:])
    n = sum(L)
    count = lambda n: len([1 for x in L if x == n])
    S = list(Set(L))
    d = _reduce(lambda x, y: x*y, (_factorial(count(z)) for z in S))
    return binom(n, L) // d


# Counts the number of edges in a subgraph of the complete graph determined by L
def _binom_sum(L):
    return sum(_binomial(z, 2) for z in L)


def _Poincare(L):
    """
    Construct the Poincaré polynomial (in Y) of the braid arrangement
    in ``|L|-1`` affine space.

    EXAMPLES::

        sage: from hypigu.Braid import _Poincare
        sage: _Poincare(4)
    """
    from sage.all import var
    Y = var('Y')
    factors = (1 + z * Y for z in range(1, len(L)))
    return _reduce(lambda x, y: x * y, factors, 1)

    # better code using polynomials
    # from sage.all import PolynomialRing, ZZ
    # A = PolynomialRing(ZZ, 'Y')
    # return A.prod(A([1, z]) for z in range(1, len(L)))


# Constructs the Igusa integral for the braid arrangement with little repetition
# of work.
def _recursive_crank(p, t, n, known=[], style="standard"):
    from sage.all import Partitions
    if not known:
        known = [_Igusa_braid_table(p, t, k, style=style)
            for k in range(_TABLE_CUTOFF + 1)]
    k = len(known) + 1
    Zk = 0
    for L in Partitions(k):
        if len(L) > 1:
            if style == "reduced":
                L_factors = [_P(L), 1, t**(_binom_sum(L)), _factorial(len(L))]
            else:
                L_factors = [
                    _P(L), p**(1 - _reduce(lambda x, y: x + y - 1, L)),
                    t**(_binom_sum(L)), _Poincare(L)(Y=-p**-1)
                ]
            lower_integrals = list(map(lambda z: known[z - 1], list(L)))
            L_term = _reduce(lambda x, y: x*y, L_factors + lower_integrals, 1)
            Zk += L_term
    if style == "reduced":
        Zk = Zk/(1 - t**(_binomial(k, 2)))
    else:
        Zk = Zk/(1 - p**(-k + 1)*t**(_binomial(k, 2)))
    known += [Zk]
    if k - 1 < n:
        return _recursive_crank(p, t, n, known=known, style=style)
    else:
        return known[n]


def BraidArrangementIgusa(n):
    r"""
    Return the rational function associated to the local Igusa zeta function for
    the n-dimensional essential braid arrangement.

    INPUT:

    - ``n`` -- integer; the dimension of the ambient affine space.

    OUTPUT: A rational function in at most two variables.

    EXAMPLES:

    This example illustrates what to expect ::

        sage: Z = BraidArrangementIgusa(2)
        sage: Z
        -(2*t/q - 2/q - t/q^2 + 1)*(1/q - 1)/((t^3/q^2 - 1)*(t/q - 1))
    """
    from sage.all import var
    p = var('q')
    t = var('t')
    if n <= _TABLE_CUTOFF:
        return _Igusa_braid_table(p, t, n, style="standard")
    return _recursive_crank(p, t, n, style="standard")



#
#   Copyright 2020 Joshua Maglione
#
#   Distributed under MIT License
#

def _non_Weyl_arrangements(X, n, shift=[0]):
    from functools import reduce
    from sage.all import HyperplaneArrangements, QQ, CoxeterGroup

    def add_shift(x):
        return list(map(lambda k: [k] + list(x), shift))
    if X == 'I':
        # Not expecting n <= 2 for type I.
        if n == 3:
            return _arrangements_from_roots("A", 2, shift=shift)
        if n == 4:
            return _arrangements_from_roots("B", 2, shift=shift)
        H = HyperplaneArrangements(QQ, tuple(['x0', 'x1']))
        norms = [tuple([1, 0]), tuple([0, 1]), tuple([1, 1]), tuple([1, -1])]
        cval = 2
        for k in range(n - 4):
            norms.append(tuple([(-1)**(k % 2), cval + (k // 2)]))
    else:
        if n == 2:
            return _non_Weyl_arrangements("I", 5, shift=shift)
        W = CoxeterGroup([X, n])
        H = HyperplaneArrangements(W.base_ring(), tuple(['x' + str(k) for k in range(n)]))
        norms = W.positive_roots()
    aff_norms = reduce(lambda x, y: x+y, map(lambda x: add_shift(x), norms), [])
    return H(aff_norms)


def _arrangements_from_roots(X, n, shift=[0]):
    from functools import reduce
    from sage.all import HyperplaneArrangements, QQ, RootSystem
    Phi = RootSystem([X, n]).ambient_space()
    d = Phi.dimension()
    H = HyperplaneArrangements(QQ, tuple(['x' + str(i) for i in range(d)]))
    norms = Phi.positive_roots()

    def add_shift(x):
        norm = tuple(map(lambda i: QQ(i), x._vector_().list()))
        return list(map(lambda k: [norm, k], shift))
    aff_norms = reduce(lambda x, y: x+y, map(lambda x: add_shift(x), norms), [])
    return H(aff_norms)


# Verifies that the Coxeter-theoretic data is expected.
def _Coxeter_check(X, n):
    if n <= 0:
        raise ValueError("Expected a positive rank.")
    if X not in ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']:
        raise ValueError(
            "Expected first character to be in {A, B, C, D, E, F, G, H, I}."
        )
    if X in ['E', 'F', 'G', 'H']:
        if X == 'E' and n not in {6, 7, 8}:
            raise ValueError(
                "Expected rank to be either 6, 7, or 8."
            )
        if X == 'F' and n != 4:
            raise ValueError(
                "Expected rank to be 4."
            )
        if X == 'G' and n != 2:
            raise ValueError(
                "Expected rank to be 2."
            )
        if X == 'H' and n not in {2, 3, 4}:
            raise ValueError(
                "Expected rank to be either 2, 3, or 4."
            )
        if X == 'I' and n <= 2:
            raise ValueError(
                "Expected rank above 2."
            )
    return True


# Parses the Coxeter type input and runs the check for good input.
def _parse_Coxeter_input(name):
    if isinstance(name, str):
        factors = name.split(' ')
    else:
        factors = list(name)
        if not isinstance(factors[0], str):
            raise TypeError("Expected ``name`` to be an interable container of strings.")

    def convert(s):
        if len(s) <= 1:
            raise ValueError("Cannot parse input: {0}".format(s))
        try:
            n = int(s[1:])
        except ValueError:
            raise ValueError("Cannot parse as integer: {0}".format(s[1:]))
        return tuple([s[0].upper(), n])
    Cox_facts = list(map(convert, factors))
    assert all(map(lambda X: _Coxeter_check(X[0], X[1]), Cox_facts))
    return Cox_facts


# Builds the direct sum arrangement of arrangements A and B.
def _direct_sum(A, B):
    from sage.all import HyperplaneArrangements as HA
    K = A.base_ring()
    d = A.dimension() + B.dimension()
    A_H = map(lambda H: H.coefficients(), A.hyperplanes())
    B_H = map(lambda H: H.coefficients(), B.hyperplanes())
    HH = HA(K, tuple(['x' + str(k) for k in range(d)]))
    A_emb = map(lambda L: L + [0]*(B.dimension()), A_H)
    B_emb = map(lambda L: L[0:1] + [0]*(A.dimension()) + L[1:], B_H)
    return HH(list(A_emb) + list(B_emb))


def _basic_wrapper(name, s):
    factors = _parse_Coxeter_input(name)

    def irr_facts(X):
        if X[0] in {'I', 'H'}:
            return _non_Weyl_arrangements(X[0], X[1], shift=s)
        if X == ['D', 1]:
            X = ['A', 1]
        return _arrangements_from_roots(X[0], X[1], shift=s)
    irr_arr = list(map(irr_facts, factors))
    return DirectSum(irr_arr)


def _res_arr(n):
    from sage.all import Subsets, HyperplaneArrangements, QQ, Matrix
    Subs = Subsets(range(1, n + 1))

    def S_vec(S):
        v = [QQ(0)]*(n+1)
        for s in S:
            v[s] = QQ(1)
        return v
    M = [S_vec(S) for S in Subs if len(S) > 0]
    HH = HyperplaneArrangements(QQ, tuple(['x' + str(k) for k in range(n)]))
    H = HH(Matrix(QQ, M))
    return H


def DirectSum(*args):
    r"""
    Return the direct sum of hyperplane arrangements.

    INPUT:

    - An iterable container of hyperplane arrangements.

    OUTPUT: the direct sum arrangement given as a hyperplane arrangement.

    EXAMPLE:

        sage: A = hi.CoxeterArrangement("A1")
        sage: Bool3 = hi.DirectSum(A, A, A)
        sage: Bool3
        Arrangement <x4 - x5 | x2 - x3 | x0 - x1>
        sage: Bool64 = hi.DirectSum([A]*64)
        sage: Bool64
        Arrangement of 64 hyperplanes of dimension 128 and rank 64
    """

    from functools import reduce
    if len(args) == 1:
        HPAs = args[0]
    else:
        HPAs = args
    if len(HPAs) == 1:
        return HPAs[0]
    else:
        D = _direct_sum(HPAs[0], HPAs[1])
        return reduce(lambda A, B: _direct_sum(A, B), HPAs[2:], D)


def CoxeterArrangement(name):
    r"""
    Return the Coxeter arrangement of the prescribed Coxeter type.

    INPUT:

    - ``name`` -- an interable container of strings; the Coxeter group
        name. The string must include a letter from {A, ..., H} and a
        nonnegative integer.

    OUTPUT: the Coxeter arrangement given as a hyperplane arrangement.

    EXAMPLES:

        sage: hi.CoxeterArrangement("B4")
        Arrangement of 16 hyperplanes of dimension 4 and rank 4

        sage: hi.CoxeterArrangement("A1 A2")
        Arrangement <x3 - x4 | x2 - x3 | x2 - x4 | x0 - x1>

        sage: hi.CoxeterArrangement(["D4", "E6"])
        Arrangement of 48 hyperplanes of dimension 12 and rank 10
    """
    return _basic_wrapper(name, [0])


def ShiArrangement(name):
    r"""
    Return the Shi arrangement of the Coxeter prescribed type.

    INPUT:

    - ``name`` -- an interable container of strings; the Coxeter group
        name. The string must include a letter from {A, ..., H} and a
        nonnegative integer.

    OUTPUT: the Shi arrangement given as a hyperplane arrangement.

    EXAMPLES:

        sage: hi.ShiArrangement("A2")
        Arrangement of 6 hyperplanes of dimension 3 and rank 2

        sage: hi.ShiArrangement("A1 A1")
        Arrangement <x2 - x3 | x2 - x3 + 1 | x0 - x1 | x0 - x1 + 1>

        sage: hi.ShiArrangement(["H3", "B3"])
        Arrangement of 48 hyperplanes of dimension 6 and rank 6
    """
    return _basic_wrapper(name, [0, 1])


def LinialArrangement(name):
    r"""
    Return the Linial arrangement of the prescribed Coxeter type.

    INPUT:

    - ``name`` -- an interable container of strings; the Coxeter group
        name. The string must include a letter from {A, ..., H} and a
        nonnegative integer.

    OUTPUT: the Linial arrangement given as a hyperplane arrangement.

    EXAMPLES:

        sage: hi.LinialArrangement("D4")
        Arrangement of 12 hyperplanes of dimension 4 and rank 4

        sage: hi.LinialArrangement("A1 A1")
        Arrangement <x2 - x3 + 1 | x0 - x1 + 1>

        sage: hi.LinialArrangement(["I4", "F4"])
        Arrangement of 28 hyperplanes of dimension 6 and rank 6
    """
    return _basic_wrapper(name, [1])


def CatalanArrangement(name):
    r"""
    Return the Catalan arrangement of the prescribed Coxeter type.

    INPUT:

    - ``name`` -- an interable container of strings; the Coxeter group
        name. The string must include a letter from {A, ..., H} and a
        nonnegative integer.

    OUTPUT: the Catalan arrangement given as a hyperplane arrangement.

    EXAMPLES:

        sage: hi.CatalanArrangement("B4")
        Arrangement of 48 hyperplanes of dimension 4 and rank 4

        sage: hi.CatalanArrangement("A1 A1")
        Arrangement of 6 hyperplanes of dimension 4 and rank 2

        sage: hi.CatalanArrangement(["D4", "A3"])
        Arrangement of 54 hyperplanes of dimension 8 and rank 7
    """
    return _basic_wrapper(name, [-1, 0, 1])


def ResonanceArrangement(n):
    r"""
    Return the rank-n resonance arrangement.

    INPUT:

    - ``n`` -- a positive integer.

    OUTPUT: the resonance arrangement given as a hyperplane arrangement.

    EXAMPLES:

        sage: hi.ResonanceArrangement(3)
        Arrangement of 7 hyperplanes of dimension 3 and rank 3

        sage: hi.ResonanceArrangement(2)
        Arrangement <x1 | x0 | x0 + x1>

        sage: hi.ResonanceArrangement(10)
        Arrangement of 1023 hyperplanes of dimension 10 and rank 10
    """
    return _res_arr(n)


def PolynomialToArrangement(f):
    r"""
    Return the associated hyperplane arrangement of f, the given product of
    linear polynomials.

    INPUT:

    - ``f`` -- a polynomial or symbolic expression.

    OUTPUT: the hyperplane arrangement associated with the linear factors of f.

    EXAMPLES:

        sage: f = SR('X^3*Y^2*Z')
        sage: f
        X^3*Y^2*Z
        sage: hi.PolynomialToArrangement(f)
        Arrangement <Z | Y | X>

        sage: f = 'X*Y*Z*W*(X - Y)*(X - Z)*(Y - W)'
        sage: hi.PolynomialToArrangement(f)
        Arrangement of 7 hyperplanes of dimension 4 and rank 4
    """
    from .GenFunctions import _parse_poly
    A, _ = _parse_poly(f)
    return A



#
#   Copyright 2021--2023 Joshua Maglione 
#
#   Distributed under MIT License
#

__version__ = 1.4

from .src.Braid import BraidArrangementIgusa
from .src.Constructors import CoxeterArrangement, LinialArrangement, ShiArrangement, CatalanArrangement, DirectSum, PolynomialToArrangement, ResonanceArrangement
from .src.LatticeFlats import LatticeOfFlats
from .src.GenFunctions import FlagHilbertPoincareSeries, IgusaZetaFunction, CoarseFlagHPSeries, AnalyticZetaFunction, AtomZetaFunction, TopologicalZetaFunction
from .src.Database import internal_database